from dios.dios import *


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data
        self._columns = _dios.columns


class _LocIndexer(_Indexer):

    def __init__(self, _dios):
        super().__init__(_dios)
        # we can use set item here, as this
        # also uses .loc for setting values
        self._set_item = _dios._set_item

    def __setitem__(self, key, val):
        keys, rkey, lowdim = self._unpack_key(key)
        ix, ixalign = self._unpack_rowkey(rkey)
        gen = _unpack_value(keys, ix, val)

        for tup in gen:
            self._set_item(*tup, ixalign=ixalign)

    def __getitem__(self, key):
        keys, rkey, lowdim = self._unpack_key(key)
        ix, ixalign = self._unpack_rowkey(rkey)

        if is_hashable(ix):
            new = pd.Series()
            new.name = ix
        else:
            new = self._dios.copy_empty()

        # set series in new dios OR set values in
        # new series if ix is hashable (see above)
        for k in keys:
            new[k] = self._get_item(self._data.loc[k], ix, ixalign=ixalign)
            maby_set_series_name(new[k], k)

        # squeeze to series if a single label was given
        # OR squeeze to val if additional ix is hashable
        if lowdim:
            new = new.squeeze()

        return new

    def _get_item(self, ser, ix, ixalign=False):
        if ixalign:
            ix = ser.index.intersection(ix.index)
        return ser.loc[ix]

    def _unpack_rowkey(self, rkey):
        align = False
        if is_dios_like(rkey) or is_nested_list_like(rkey):
            raise ValueError("Cannot index with multidimensional key")
        if is_bool_indexer(rkey):
            if not isinstance(rkey, pd.Series):
                raise ValueError("Must pass Series with boolean values only")
            rkey, align = rkey[rkey], True  # kill `False`
        return rkey, align

    def _unpack_key(self, key):
        lowdim = False
        if isinstance(key, tuple):
            key, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")
            if is_dios_like(ckey):
                raise ValueError("Cannot index with multidimensional key")
            if is_hashable(ckey):
                keys = [ckey]
                lowdim = True
            else:
                keys = self._data.loc[ckey].index.to_list()
        else:
            keys = self._columns.to_list()
        return keys, key, lowdim


class _iLocIndexer(_Indexer):

    def __getitem__(self, key):
        rkey, cols, lowdim = self._unpack_key(key)
        if is_scalar(rkey[0]):
            return self._series(rkey, cols, lowdim)
        elif lowdim:
            return self._scalar(rkey[0], cols[0])
        else:
            new = self._dios.copy_empty()
            for i, _ in enumerate(cols):
                c, r = cols[i], rkey[i]
                new[c] = self._data[c].iloc[r]
            return new

    def _series(self, rkey, cols, lowdim):
        if lowdim:
            return self._scalar(rkey[0], cols[0])
        new = pd.Series()
        for c in cols:
            try:
                new[c] = self._data[c].iloc[rkey]
            except KeyError:
                new[c] = np.nan

    def _scalar(self, r, c):
        return self._data[c].iloc[r]

    def __setitem__(self, key, value):
        ixs, keys, _ = self._unpack_key(key)
        gen = self._unpack_value(keys, ixs, value)
        for tup in gen:
            self._set_item_positional(*tup)
        raise NotImplemented

    def _set_item_positional(self, key, ix, val):
        ser = self._data[key]
        if isinstance(val, pd.Series):
            index = ser.iloc[ix].index
            index = index.intersection(val.index)
            if not index.empty:
                ser.loc[index] = val.loc[index].copy()
        else:
            ser.iloc[ix] = val

    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        lowdim = False
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            ckey = list(ckey) if is_iterator(ckey) else ckey

            # determine columns
            if is_integer(ckey):
                self._check_keys([ckey])
                cols = self._integers_to_col_list([ckey])
                lowdim = True
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            elif is_list_like(ckey) and not is_nested_list_like(ckey):
                arr = np.array(ckey)
                if is_bool_array(arr):
                    raise NotImplementedError
                self._check_keys(ckey)
                cols = self._integers_to_col_list(ckey)
            elif isinstance(ckey, pd.Series):
                raise NotImplementedError
            elif is_bool_indexer(ckey):
                raise NotImplementedError
            else:
                raise KeyError(f"{ckey} of type {type(ckey)}")
        else:
            cols = list(self._data.index)
            rkey = key

        # blowup
        rkey = [rkey] * len(cols)
        return rkey, cols, lowdim

    def _check_keys(self, keys):
        bound = len(self._data)
        for k in keys:
            if not is_integer(k):
                raise ValueError(f"{type(k)} is not integer")
            if k not in range(-bound, bound):
                raise KeyError("positional indexer(s) are out-of-bounds in columns")

    def _integers_to_col_list(self, ints):
        klist = list(self._data.index)
        ks = set()
        for i in ints:
            ks.add(klist[i])
        return list(ks)

    def _col_slice_to_col_list(self, sl):
        for s in [sl.start, sl.stop, sl.step]:
            if not is_integer(s):
                raise TypeError(f"positional indexing with slice must be integers, passed type was {type(s)}")
        return list(self._data.index)[sl]


def _unpack_value(keys, ix, val):
    """Return a generator that yield (column key, corresponding value, value-align(bool) )
    for all columns.
    This is analogous to DictOfSeries._unpack_value, but with some modifications."""

    # prepare value
    val = list(val) if is_iterator(val) else val
    val = val.squeeze() if is_dios_like(val) else val
    dioslike, nlistlike = is_dios_like(val), is_nested_list_like(val)

    # check value
    if nlistlike and len(val) != len(keys):
        raise ValueError(f"could not broadcast input array with length {len(val)}"
                         f" into dios of length {len(keys)}")
    if dioslike:
        keys = val.columns.intersection(keys)

    for i, k in enumerate(keys):
        if dioslike:
            yield k, ix, val[k]
        elif nlistlike:
            yield k, ix, val[i]
        else:
            yield k, ix, val


def maby_set_series_name(maybe_ser, name):
    if isinstance(maybe_ser, pd.Series):
        maybe_ser.name = name
